Package: scam
Version: 1.2-22
Authors@R: person(given = "Natalya",
                      family = "Pya",
                      role = c("aut", "cre"),
                      email = "nat.pya@gmail.com")
Maintainer: Natalya Pya <nat.pya@gmail.com>
Title: Shape Constrained Additive Models
Date: 2026-03-12
Description: Generalized additive models under shape
        constraints on the component functions of the linear predictor.       
        Models can include multiple shape-constrained (univariate
        and bivariate) and unconstrained terms. Routines of the 
        package 'mgcv' are used to set up the model matrix, print, 
        and plot the results. Multiple smoothing parameter 
        estimation by the Generalized Cross Validation or similar.
        See Pya and Wood (2015) <doi:10.1007/s11222-013-9448-7> 
        for an overview. A broad selection of shape-constrained 
        smoothers, linear functionals of smooths with shape constraints,
        and Gaussian models with AR1 residuals. 
Depends: R (>= 3.6.0)
Imports: mgcv (>= 1.8-2), methods, stats, graphics, Matrix, splines
Suggests: nlme, emmeans (>= 1.4.2), estimability
License: GPL (>= 2)
LazyLoad: yes
NeedsCompilation: yes
Packaged: 2026-03-12 09:20:51 UTC; natalya
Author: Natalya Pya [aut, cre]
Repository: CRAN
Date/Publication: 2026-03-12 10:10:02 UTC
